/*-------------------------------------------------------------
backtrack.c
P. Clote, Jul 2004
-------------------------------------------------------------*/




                                                                                

#ifndef __SPECTRUM_H__
#include "spectrum.h"
#endif
                                                                                
void initializeParen(short** paren,int n){
  int level,i;
  //initialize length
  for (level=0;level<n/2;level++){
     paren[level][0]   = n;    //length of string
     paren[level][n+1] = '\0'; //string terminator
     }
  //initialize characters to empty secondary structure 
  for (level=0;level<n/2;level++)
    for (i=1;i<=n;i++)
      paren[level][i] = '.';
  }


void backtrack(int i, int j, int k, int s, int b, short* paren, 
               short *a, MatrixType M){
   //a is RNA sequence a[1]....a[n] of length n = a[0]
   //M is the matrix from energy computation
   // backtrack is recursive function, where 1 <= i < j <= n
   //k is level, s is visibility
   int r,x,y,kLeft,kRight,sLeft,sRight,bLeft,bRight,s0,b0;
   if (k>=0 && i>=1 && i+THRESHOLD < j && BP(i,j,k) > 0 ){
     //NOTE: BP(i,j,k) > 0  means that there IS a k-subopt
     //sec str and that the number of basepairs in that sec str.
     if (DEBUG){
       printf("Calling backtrack, i:%d j:%d k:%d s:%d b:%d\n", i,j,k,s,b);
       printf("                   BP(%d,%d,%d) = %d\n",i,j,k,BP(i,j,k));
       }
     r       = BPpartner(i,j,k,s,b);
     kLeft   = LEVELleft(i,j,k,s,b);
     kRight  = LEVELright(i,j,k,s,b);
     bLeft   = BSleft(i,j,k,s,b);
     bRight  = BSright(i,j,k,s,b);
     sLeft   = VISleft(i,j,k,s,b);
     sRight  = VISright(i,j,k,s,b);
     if(r == 0)  {
       if (j-i-1>THRESHOLD ){
         if (DEBUG) printf("Case 3, i:%d j:%d k:%d s:%d b:%d\n", i,j-1,kLeft,sLeft,bLeft);
         backtrack(i,j-1,kLeft,sLeft,bLeft,paren,a,M);
         }
       }
     else if(r == i)  {
       paren[r] = '('; 
       paren[j] = ')';
       if (j-i-2>THRESHOLD ){
         if (DEBUG) printf("Case 1, i:%d j:%d k:%d s:%d b:%d\n", i+1,j-1,kRight,sRight,bRight);
         backtrack(i+1,j-1,kRight,sRight,bRight,paren,a,M);
         }
       }
     else { //i<r<j
       paren[r] = '('; 
       paren[j] = ')';
       x = r-i; 
       y = (j-1)-(r+1)+1; 
       //as in deterministic quicksort, we handle the 
       //shorter recursive call first -- this ensures stack depth at most log
       if (x<y) {
       if( THRESHOLD < x ){
         backtrack(i,r-1,kLeft,sLeft,bLeft,paren,a,M);
         if (DEBUG){
           printf("Case 2 (%d,%d basepair), i:%d r-1:%d k:%d s:%d b:%d \n",r,j,i,r-1,kLeft,sLeft,bLeft);
           }
         }
       if (THRESHOLD < y  ){
         backtrack(r+1,j-1,kRight,sRight,bRight,paren,a,M);
         if (DEBUG) printf("Case 2 (%d,%d basepair), r+1:%d j-1:%d k:%d s:%d b:%d \n",\
                            r,j,r+1,j-1,kRight,sRight,bRight);
         }
       }
       else { // so x >= y
       if (THRESHOLD < y  ){
         backtrack(r+1,j-1,kRight,sRight,bRight,paren,a,M);
         if (DEBUG) printf("Case 2 (%d,%d basepair), r+1:%d j-1:%d k:%d s:%d b:%d\n",
                            r,j,r+1,j-1,kRight,sRight,bRight);
         }
       if( THRESHOLD < x ){
         backtrack(i,r-1,kLeft,sLeft,bLeft,paren,a,M);
         if (DEBUG) printf("Case 2 (%d,%d basepair), i:%d r-1:%d k:%d s:%d b:%d\n",r,j,i,r-1,kLeft,sLeft,bLeft);
         }
       }
     } //end else
  } // end if ( B(i,j,k) > 0 )
}
   


void createLevelSecStr(short** paren,int maxK,short *a,MatrixType M){
  int k,n,s,b;
  n = a[0];
  for (k=0;k<=MIN(KK-1,maxK);k++){
    if ( BP(1,n,k) > 0 ){
      s = VIS(1,n,k); 
      b = BS(1,n,k); 
      backtrack(1,n,k,s,b,paren[k],a,M);
      }
    }
}
