/*---------------------------------------------------------

computeEnergy.c
P. Clote, Jul 2004

---------------------------------------------------------*/

#ifndef __SPECTRUM_H__
#include "spectrum.h"
#endif

int computeEnergy(MatrixType M, MatrixTypeDouble MM, short *a){
  //RNA sequence a1,...,an with n length of sequence
  //This function is called after initializing M
  //Function returns maxK representing the maximum level k such
  //that there is a k-suboptimal sec str on [1,n]
 int i,j,k,k0,kLeft,kRight,d,r,n,maxNumBP,flag;
 int maxK,bp,bp0,bp1,bpLeft,bpRight,b,bLeft,bRight,s,s0,sLeft,sRight,x;
 double num,numLeft,numRight;
  /*----------------------------------------------
  NOTES:
  1) In code below, kLeft resp kRight stand for 
     what I called "l" and "l prime" in the paper "An efficient
     algorithm to compute RNA energy spectrum", and stands
     for the level of the left subregion i...r-1 resp. right subregion
     r+1...j-1 after basepairing r,j.
     Similarly, bpLeft resp bpRight represent the number of basepairs
     in the kLeft-suboptimal left subregion and the 
     kRight-suboptimal right subregion. 
  ----------------------------------------------*/

 n = a[0];
 for (d=THRESHOLD+1;d<=n-1;d++) 
  for (i=1;i<=n-THRESHOLD-1;i++){
   j = i+d;
   if (j>n) break;
   maxNumBP = BP(i,j,0); 
     //num basepairs in optimal sec str on [i,j]
   maxK     = 0;         
     //We will calculate MAXK[i,j] = maxK = argmax_k [ BPs(i,j,k,s,b)>=0, over all s,b]
   for (r=i;r<=j-THRESHOLD-1;r++){
    if ( basePair(a[r],a[j]) ){

/*--------------------------------------------------------

CASE 1
Here j is paired with i

--------------------------------------------------------*/

     if (i==r){
      for (kRight=0;kRight<=MIN(KK-1,(j-i-2)/2);kRight++){
       bpRight = BP(i+1,j-1,kRight);
       flag = ( bpRight == -1 || (kRight>0 && kRight >=BP(i+1,j-1,0))  || \
                kRight >= KK || kRight >MAXK(i+1,j-1) );
       if (flag) continue;
         //there is NO k-subopt sec str on [r+1,j-1]
       bp    = 1 + bpRight;
       k     = maxNumBP - bp; if (k==0 || k>=KK) continue;
       for (sRight=0;sRight<SS;sRight++)
        for (bRight=0;bRight<BB;bRight++)
         if (BPs(i+1,j-1,kRight,sRight,bRight) == BP(i+1,j-1,kRight)){
          s = 0; //no nucleotide is external to basepairs in [i,j]
          b = 0;
          if (k>maxK)  maxK = k;  //update maxK
          BP(i,j,k)         = bp;
          VIS(i,j,k)        = s;
          BS(i,j,k)         = b;
          num = NUMSUBOPT(i+1,j-1,kRight,sRight,bRight); 
          NUMSUBOPT(i,j,k,s,b) += num;
          BPs(i,j,k,s,b)    = bp;
          BPpartner(i,j,k,s,b) = i;
          LEVELleft(i,j,k,s,b) = 0;
          LEVELright(i,j,k,s,b) = kRight;
          VISleft(i,j,k,s,b)  = 0;
          VISright(i,j,k,s,b)  = sRight;
          BSleft(i,j,k,s,b)   = 0;
          BSright(i,j,k,s,b)  = bRight;
          } //end if (BPs(i+1,j-1,kRight,sRight,bRight) ....
       } //end for (kRight=0;kRight<=(j-i-2)/2;kRight++)
      } //end if (r==i)


/*--------------------------------------------------------

CASE 2
Here j is paired with r, i<r<j

--------------------------------------------------------*/


     else { //i<r<j
      for (kLeft=0;kLeft<=MIN(KK-1,(r-1-i)/2);kLeft++){
       bpLeft = BP(i,r-1,kLeft);
       flag = ( bpLeft == -1 || (kLeft>0 && kLeft >=BP(i,r-1,0))  || \
                kLeft >= KK || kLeft >MAXK(i,r-1) ); 
       if (flag) continue;
         //there is NO kLeft-subopt sec str on [i,r-1]
         //The second term in OR statement was added only after hours of debugging
       for (kRight=0;kRight<=MIN(KK-1,(j-i-2)/2);kRight++){
        bpRight = BP(r+1,j-1,kRight);
       flag = ( bpRight == -1 || (kRight>0 && kRight >=BP(r+1,j-1,0))  || \
                kRight>= KK || kRight >MAXK(r+1,j-1) );
       if (flag) continue;
          //there is NO kRight-subopt sec str on [r+1,j-1]
          //The second term in OR statement was added only after hours of debugging
        bp = 1 + bpLeft + bpRight;
        k    = maxNumBP - bp; if (k==0 || k>= KK) continue;
        for (sLeft=0;sLeft<SS;sLeft++)
         for (bLeft=0;bLeft<BB;bLeft++){
          bp0 = BPs(i,r-1,kLeft,sLeft,bLeft);
          if (bp0 == -1 || bp0 != BP(i,r-1,kLeft)) continue;
          for (sRight=0;sRight<SS;sRight++)
           for (bRight=0;bRight<BB;bRight++){
            bp1 = BPs(r+1,j-1,kRight,sRight,bRight);
            if (bp1 == -1 || bp1!=BP(r+1,j-1,kRight) ) continue;
            s            = joinSandB(a,sLeft,bLeft,i,r-1);
            b            = 0;
            if (k>maxK) maxK = k;  //update maxK
            BP(i,j,k)    = bp;
            VIS(i,j,k)   = s;
            BS(i,j,k)    = b;
            numLeft  = NUMSUBOPT(i,r-1,kLeft,sLeft,bLeft);
	    numRight = NUMSUBOPT(r+1,j-1,kRight,sRight,bRight); 
	    if (numLeft==0 && numRight==0)
              num = 1;		     
	    else if (numLeft==0)
              num = numRight;		     
	    else if (numRight==0)
              num = numLeft;		     
	    else
              num = numLeft*numRight;
            NUMSUBOPT(i,j,k,s,b) += num;
            BPs(i,j,k,s,b)         = bp;
            BPpartner(i,j,k,s,b)   = r;
            LEVELleft(i,j,k,s,b)   = kLeft;
            LEVELright(i,j,k,s,b)  = kRight;
            VISleft(i,j,k,s,b)     = sLeft;
            VISright(i,j,k,s,b)    = sRight;
            BSleft(i,j,k,s,b)      = bLeft;
            BSright(i,j,k,s,b)     = bRight;
            } //end for (bRight=0;bRight<BB;bRight++)
          }// end for (bLeft=0;bLeft<BB;bLeft++)
        }// end for (kRight=0;kRight<=(j-i-2)/2;kRight++)
       }// end for (kLeft=0;kLeft<=(r-1-i)/2;kLeft++)
      } //end else { // i<r<j
   } //end if ( basePair(a[r],a[j]) )
  } // end for (r=i;r<=j-THRESHOLD-1;r++)

/*--------------------------------------------------------

CASE 3  
Here j is not paired with any nucleotide in [i,j]

--------------------------------------------------------*/
   for (kLeft=0;kLeft<=(j-1-i)/2;kLeft++){
    bp = BP(i,j-1,kLeft);
       flag = ( bp == -1 || (kLeft>0 && kLeft >=BP(i,j-1,0))  || \
                kLeft >= KK || kLeft >MAXK(i,j-1) );
       if (flag) continue;
      //there is NO k-subopt sec str on [i,j-1]
    k  = maxNumBP - bp; if (k==0 || k>=KK) continue;
    for (sLeft=0;sLeft<SS;sLeft++)
     for (bLeft=0;bLeft<BB;bLeft++){
      bp0 =  BPs(i,j-1,kLeft,sLeft,bLeft);
      if ( bp0== -1  || bp!=bp0 ) continue;
      x = a[j-1-THRESHOLD];
      if ( THRESHOLD+1==bLeft && !in(x,sLeft) )
        s = addElement(x,sLeft);
      else
        s = sLeft;
      if (canBasePairWithElementInSet(a[j],s)) continue; 
          // j cannot be basepaired with visible nucleotide, i.e. from s
      b = MIN(bLeft+1,THRESHOLD+1);
      if (k>maxK) maxK = k; //update maxK
      BP(i,j,k)    = bp;
      VIS(i,j,k)   = s;
      BS(i,j,k)    = b;
      NUMSUBOPT(i,j,k,s,b) += NUMSUBOPT(i,j-1,kLeft,sLeft,bLeft);
      BPs(i,j,k,s,b)        = bp;
      BPpartner(i,j,k,s,b)  = 0; //no basepair partner of j in [i,j]
      LEVELleft(i,j,k,s,b)  = kLeft;
      LEVELright(i,j,k,s,b) = 0;
      VISleft(i,j,k,s,b)    = sLeft;
      VISright(i,j,k,s,b)   = 0;
      BSleft(i,j,k,s,b)     = bLeft;
      BSright(i,j,k,s,b)    = 0;
      }// end for (bLeft=0;bLeft<BB;bLeft++)
    } // for (kLeft=0;kLeft<=(j-1-i)/2;kLeft++)
  MAXK(i,j) = maxK;
 }// for i
 maxNumBP = BP(1,n,0);
 maxK = 0;
 for (k=0;k<MIN(KK,maxNumBP);k++)
  if ( BP(1,n,k)>0 ) maxK = k;
 return maxK;
}
