#! /usr/bin/python
# Program: generateWithDinuclFreq.py
# P. Clote

#-------------------------------------------------------------
# Program computes mono- and dinucleotide frequencies.
# By appropriate modification, one can produce a script which
# computes the mono- and dinucleotide frequencies of a collection
# of tRNA (versus rRNA), where each RNA sequence appears in a
# separate file in a given directory.
# Finally, new sequences are generated from these frequencies.
#-------------------------------------------------------------



import sys, string,whrandom


def computeFreq(file):
  #Global variables
  #Initialize mono- and dinucleotide dictionaries
  #nuclList = ["A","C","G","T"]
  nuclList   = ["A","C","G","U"]
  nuclCnt    = {}  #empty dictionary
  dinuclCnt  = {}  #empty dictionary
  nuclFreq   = {}  #empty dictionary
  dinuclFreq = {}  #empty dictionary
  pseudocount = .1  #fictive occurrence of each mononucleotide
  pseudocount2= .025 #fictive occurrence of each dinucleotide

     # this initialization to empty dictionary is necessary
  for x in nuclList:
     nuclCnt[x]=pseudocount
        #entry in mononucleotide dictionary
     for y in nuclList:
        word = [x,y]
        dinucl = string.join(word,"")
        dinuclCnt[dinucl]=pseudocount2
           #entry in dinucleotide dictionary
  
  genomeLength=pseudocount*len(nuclList)
  dinuclTotal = pseudocount2*len(nuclList)**2
  text = ""
  line = file.readline()
  while line:
    if line[0]==">":  # treat lines beginning with '>' as comment and skip 
      line = string.upper(file.readline())
      continue
    else: 
      text += line
      line = file.readline()

  text = text.replace("\n","")
  text = text.upper()

  for key in nuclCnt.keys(): 
     num = string.count(text,key)
     nuclCnt[key] = nuclCnt[key]+num
     genomeLength = genomeLength + num #compute mononucleotide abs freq
  for key in dinuclCnt.keys(): 
     num = string.count(text,key)
     dinuclCnt[key] = dinuclCnt[key]+num
     dinuclTotal = dinuclTotal + num
  
  print "Genome length: %d" % genomeLength
  print "Frequency analysis of File %s" % sys.argv[1]
  print "---------------------------------------------"
  print "Nucleotide\tCount\tRel freq"
  keys = nuclCnt.keys()
  keys.sort()
  for key in keys:
     nuclFreq[key] = float(nuclCnt[key])/float(genomeLength)
     print "%s\t\t%d\t%f" % (key,nuclCnt[key],nuclFreq[key])
  print "\n\n"
  print "Number of dinucleotides: %s" % dinuclTotal
  keys = nuclCnt.keys()
  keys.sort()
  
  print "\t\tA \t\tC \t\tG \t\tT"
  sys.stdout.write("  ")
  for i in range(70):
     sys.stdout.write("-")
  print

  sum = 0.0  	
  for key1 in nuclList:
     print "%s|\t\t" % key1,
     for key2 in nuclList:
        word=[key1,key2]
        dinuclkey = string.join(word,"")
        x= float(dinuclCnt[dinuclkey])/float(dinuclTotal)
        dinuclFreq[dinuclkey] = x
        sum += x
        print "%6.5f\t\t" % dinuclFreq[dinuclkey],
     print
  print
  #print "Check step: %f" % sum
  return nuclFreq,dinuclFreq

 
def generateSequences(nuclFreq,dinuclFreq,n,m):
  for i in range(m):
    print "> %d" % (i+1)
    x = whrandom.random()
    a=nuclFreq['A']
    c=nuclFreq['C']
    g=nuclFreq['G']
    u=nuclFreq['U']
    if x<a:
      ch = 'A' 
    elif x<a+c:
      ch = 'C' 
    elif x<a+c+g:
      ch = 'G'
    else:
      ch = 'U'
    sys.stdout.write(ch)
    for j in range(n-1):
      a=dinuclFreq[ch+'A']
      c=dinuclFreq[ch+'C']
      g=dinuclFreq[ch+'G']
      u=dinuclFreq[ch+'U']
      denom = (a+c+g+u)
      a = a/denom
      c = c/denom
      g = g/denom
      u = u/denom
      #NOTE: a+c+g+u = 1.0 These are now conditional prob Pr[A|ch] etc.
      x = whrandom.random()
      #print x,a,c,g,u,(a+c+g+u)
      if x<a:
        ch1 = 'A' 
      elif x<a+c:
        ch1 = 'C' 
      elif x<a+c+g:
        ch1 = 'G'
      else:
        ch1 = 'U'
      sys.stdout.write(ch1)
      ch = ch1
    print
  return  
 
  
  
if __name__ == '__main__':  
  if len(sys.argv) < 3 :
     print "Script to compute nucleotide frequencies."
     print "Usage: %s filename n [m]" %  sys.argv[0]
     sys.exit(1)
  file = open(sys.argv[1],"r")
  n    = int(sys.argv[2])
  if len(sys.argv)==4:
    m = int(sys.argv[3])
  else:
    m = 1
  monoFreq,diFreq = computeFreq(file)
  generateSequences(monoFreq,diFreq,n,m)
  
