/*---------------------------------------------------------

initialize.c
P. Clote, Jul 2004

---------------------------------------------------------*/

#ifndef __SPECTRUM_H__
#include "spectrum.h"
#endif


void initialize(MatrixType M, MatrixTypeDouble MM, short *a){
  //RNA sequence a1,...,an with n length of sequence
  //Function computes Nussinov Jacobson value in matrix
  //        BP(i,j,0), BP(i,j,0,s,b) for 0 =s<16, 0<=b<=(THRESHOLD+1)
  //and
  //        BPpartner(i,j,0,s,b) for 0<=s<16, 0<=b<=(THRESHOLD+1)
  int i,j,k,d,max,r0,r,bp,n,s,b;
    /*-----------------------------------
    1) i,j,k loop control variables
       By k-suboptimal sec str on [i,j], we mean a sec str on a_i,...,a_j
       having bp number of basepairs, where bp+k equals the maximum number
       of basepairs on a_i,...,a_j.

       0<=k<n/2 is the level, ie k=0 corresponds to ordinary Nussinov-Jacobson
       optimal sec str, while k>0 corresponds to k-suboptimal sec str
       For k=0 and i<=j, BP(i,j,k) is the num basepairs in optimal
       sec structure on [i,j].  In particular, BP(i,j,0) is 0 when no
       basepairing is available.

       The same is true for k>0 and k-suboptimal sec str; however, instead of 
       initializing BP(i,j,k) to 0, in the case k>0 we initialize to -1,
       which will represent that there is NO k-suboptimal sec str in [i,j],
       as opposed to representing that the k-suboptimal sec str has k fewer 
       basepairs than the optimal.
     
       
    2) d distance from principal diagonal in dynamic programming algorithm
       to fill in matrix M.
    3) r0 is value r, such that r and j basepair and for which the max
       number of basepairs in [i,j] is obtained by basepairing r,j;
       otherwise r0 is 0, to indicate that j does not basepair with
       any index in interval [i,j] in obtaining the optimal number of basepairs.
    4) bp stands for number of basepairs
    -------------------------------------*/
  n = a[0]; //length of RNA sequence a[1],...,a[n]

#if 0
  //initialize to 0
  for (i=1;i<=n-1;i++)
    for (j=i;j<=n;j++)
      for (k=0;k<KK;k++)
        for (s=0;s<SS0;s++)
          for (b=0;b<BB0;b++)
            M[A4*i+A3*j+A2*k+A1*s+b] = 0;
  for (i=1;i<=n-1;i++)
    for (j=i;j<=n;j++)
      for (k=0;k<KK;k++)
        for (s=0;s<SS;s++)
          for (b=0;b<BB;b++)
            MM[B4*i+B3*j+B2*k+B1*s+b] = 0;
#endif


  //First, set ALL values
  for (i=1;i<=n-1;i++)
    for (j=i;j<=n;j++){
      MAXK(i,j) = 0;
      for (k=0;k<KK;k++){
        BP(i,j,k)  = -1;
        s          = initializeVis(a,i,j);
        b          = initializeBitString(i,j); 
        VIS(i,j,k) = s;
        BS(i,j,k)  = b;
        for (s=0;s<SS;s++)
          for (b=0;b<BB;b++){
            NUMSUBOPT(i,j,k,s,b)  =  0;
            BPs(i,j,k,s,b)        = -1;
            BPpartner(i,j,k,s,b)  = -1;
            LEVELleft(i,j,k,s,b)  = -1;
            LEVELright(i,j,k,s,b) = -1;
            VISleft(i,j,k,s,b)    = initializeVis(a,i,j-1);
            VISright(i,j,k,s,b)   = 0;
            BSleft(i,j,k,s,b)     = initializeBitString(i,j-1);
            BSright(i,j,k,s,b)    = 0;
            }//for b
        }//for k
     }//for j




  //Now set nontrivial values
  for (i=1;i<=n;i++)
    for (j=i;j<=n;j++){
      b          = initializeBitString(i,j); 
      s          = initializeVis(a,i,j);
      BP(i,j,0)  = 0;
      BS(i,j,0)  = b;
      VIS(i,j,0) = s;
      if ( j-i<= THRESHOLD )
        NUMSUBOPT(i,j,0,s,b)  = 1; //empty structure only
      BPs(i,j,0,s,b)        = 0;
      BPpartner(i,j,0,s,b)  = 0;
      LEVELleft(i,j,0,s,b)  = 0;
      LEVELright(i,j,0,s,b) = 0;
      VISleft(i,j,0,s,b)    = initializeVis(a,i,j-1);
      VISright(i,j,0,s,b)   = 0;
      BSleft(i,j,0,s,b)     = initializeBitString(i,j-1); 
      BSright(i,j,0,s,b)    = 0;
      } //end for i,j



  for (d=THRESHOLD+1;d<=n-1;d++)
    for (i=1;i<=n-THRESHOLD-1;i++){
      j = i+d;
      if (j>n) break;
      max = BP(i,j-1,0);
      r0  = 0;
      for (r=i;r<=j-THRESHOLD-1;r++){
        if ( basePair(a[r],a[j]) ){
          if (r==i){
            bp = 1+BP(r+1,j-1,0);
            }
          else {
            bp = 1+BP(i,r-1,0)+BP(r+1,j-1,0);
            }
          if ( bp>max ){ max = bp; r0  = r; }
          }
        }
      BP(i,j,0) = max;
    }
  //Now have computed BP(i,j,0), which for fixed i,j is later called maxNumBP
}


