/*---------------------------------------------------------

initializeBis.c
P. Clote, Jul 2004

---------------------------------------------------------*/

#ifndef __SPECTRUM_H__
   #include "spectrum.h"
#endif


void initializeBis(MatrixType M, MatrixTypeDouble MM, short *a){
  //RNA sequence a1,...,an with n length of sequence
  int i,j,k,d,max,r0,r,bp,bp0,bp1,n,s,sLeft,sRight,b,bLeft,bRight,x;
  int numNativeStates;
    /*-----------------------------------
    1) i,j,k loop control variables
       By k-suboptimal sec str on [i,j], we mean a sec str on a_i,...,a_j
       having bp number of basepairs, where bp+k equals the maximum number
       of basepairs on a_i,...,a_j.

       0<=k<n/2 is the level, ie k=0 corresponds to ordinary Nussinov-Jacobson
       optimal sec str, while k>0 corresponds to k-suboptimal sec str
       For k=0 and i<=j, BP(i,j,k) is the num basepairs in optimal
       sec structure on [i,j].  In particular, BP(i,j,k) is 0 when no
       basepairing is available. 

       The same is true for k>0 and k-suboptimal sec str; however, instead of 
       initializing BPs(i,j,k,s) to 0, in the case k>0 we initialize to -1,
       which will represent that there is NO k-suboptimal sec str in [i,j],
       as opposed to representing that the k-suboptimal sec str has k fewer 
       basepairs than the optimal.
       BP(i,j,k)  = max{ BPs(i,j,k,s) : 0<=s<16}
     
       For s=0,...,15, BPs(i,j,k,s) is the num basepairs in  s-visible
       k-suboptimal sec str S on a[i],...,a[j]. In other words, nucleotides 
       in s are external to any basepair of S, where s is subset of 
       {A,C,G,U}, (really s codes such a set):
            BIT(s,0) is 1/0 according to whether A in S,
            BIT(s,1) is 1/0 according to whether C in S,
            BIT(s,2) is 1/0 according to whether G in S,
            BIT(s,3) is 1/0 according to whether U in S.

       
    2) d distance from principal diagonal in dynamic programming algorithm
       to fill in matrix M.
    3) r0 is value r, such that r and j basepair and for which the max
       number of basepairs in [i,j] is obtained by basepairing r,j;
       otherwise r0 is 0, to indicate that j does not basepair with
       any index in interval [i,j] in obtaining the optimal number of basepairs.
    4) bp stands for number of basepairs
    -------------------------------------*/


  n  = a[0];

  for (d=THRESHOLD+1;d<=n-1;d++)
    for (i=1;i<=n-THRESHOLD-1;i++){
      j = i+d;
      if (j>n) break;
      for (r=i;r<=j-THRESHOLD-1;r++){
        if ( basePair(a[r],a[j]) ){


/*-----------------------------------------------------------
CASE 1
In this case, j is pairs with i
-----------------------------------------------------------*/

          if (r==i){
            bp = 1+BP(i+1,j-1,0);
            if (  bp==BP(i,j,0)  ){
              for (sRight=0;sRight<SS;sRight++)
                for (bRight=0;bRight<BB;bRight++)
                  if ( BPs(i+1,j-1,0,sRight,bRight) == BP(i+1,j-1,0) ){
                    s = 0; //no nucleotide is external to basepairs in [i,j]
                    b = 0;
                    VIS(i,j,0)            = s;
                    BS(i,j,0)             = b;
                    NUMSUBOPT(i,j,0,s,b) += NUMSUBOPT(i+1,j-1,0,sRight,bRight); 
                    BPs(i,j,0,s,b)        = bp;
                    BPpartner(i,j,0,s,b)  = i;
                    LEVELleft(i,j,0,s,b)  = 0;
                    LEVELright(i,j,0,s,b) = 0;
                    VISleft(i,j,0,s,b)    = 0;
                    VISright(i,j,0,s,b)   = sRight;
                    BSleft(i,j,0,s,b)     = 0;
                    BSright(i,j,0,s,b)    = bRight;
                    } //end if ( BPs(i+1,j-1,0,sRight,bRight) == BP(i+1,j-1,0) )
              } //end if (  bp==BP(i,j,0)  )
            } //end if (r==i)



/*-----------------------------------------------------------
CASE 2
In this case, j is pairs with r, for i<r<j.
-----------------------------------------------------------*/


          else { //i<r<j
            bp = 1+BP(i,r-1,0)+BP(r+1,j-1,0);
            if (  bp==BP(i,j,0)  ){
              for (sLeft=0;sLeft<SS;sLeft++)
                for (bLeft=0;bLeft<BB;bLeft++){
                  bp0 = BPs(i,r-1,0,sLeft,bLeft);
                  if (bp0 == -1 || bp0 != BP(i,r-1,0)) continue;
                  for (sRight=0;sRight<SS;sRight++)
                    for (bRight=0;bRight<BB;bRight++){
                      bp1 = BPs(r+1,j-1,0,sRight,bRight);
                      if (bp1 == -1 || bp1 != BP(r+1,j-1,0)) continue;
                      s  = joinSandB(a,sLeft,bLeft,i,r-1);
                      b                       = 0;
                      VIS(i,j,0)              = s;
                      BS(i,j,0)               = b;
                      BPs(i,j,0,s,b)          = bp;
                      NUMSUBOPT(i,j,0,s,b) += \
       (NUMSUBOPT(i,r-1,0,sLeft,bLeft) * NUMSUBOPT(r+1,j-1,0,sRight,bRight)); 
                      BPpartner(i,j,0,s,b)    = r;
                      LEVELleft(i,j,0,s,b)    = 0;
                      LEVELright(i,j,0,s,b)   = 0;
                      VISleft(i,j,0,s,b)      = sLeft;
                      VISright(i,j,0,s,b)     = sRight;
                      BSleft(i,j,0,s,b)       = bLeft;
                      BSright(i,j,0,s,b)      = bRight;
                      //NOTE: The previous makes sense even if BP(i,r-1,0)=0
                        //In this case, the only way that BPs(i,r-1,0,s) = 0
                        //(instead of -1) is if s = nuclAppears(a,i,r-1)
                      }// for (bRight=0;bRight<BB;bRight++)
                  }//end for (bLeft=0;bLeft<BB;bLeft++)
              } //end if (  bp==BP(i,j,0)  )
            }  //end else { // i<r<j
          } //end if ( basePair(a[r],a[j]) )
        } // end for (r=i;r<=j-THRESHOLD-1;r++)

 
/*-----------------------------------------------------------
CASE 3
In this case, j is left unpaired with any r in [i,j]
This can happen for 2 reasons: either j cannot pair with any r in [i,j]
or it can, but those nucleotides with which it can basepair are
internal to basepairs of an optimal structure on [i,j-1] which is
also optimal on [i,j].
-----------------------------------------------------------*/
      bp = BP(i,j-1,0);
      if (bp != BP(i,j,0)) continue;
      for (sLeft=0;sLeft<SS;sLeft++)
        for (bLeft=0;bLeft<BB;bLeft++){
          bp0 = BPs(i,j-1,0,sLeft,bLeft);
          x = a[j-1-THRESHOLD];
          if ( THRESHOLD+1==bLeft && !in(x,sLeft) )
            s = addElement(x,sLeft);
          else
            s = sLeft;
          if (  bp0 == bp && !canBasePairWithElementInSet(a[j],s) ){
            b = MIN(bLeft+1,THRESHOLD+1);
            VIS(i,j,0)             = s;
            BS(i,j,0)              = b;
            BPs(i,j,0,s,b)         = bp;
            NUMSUBOPT(i,j,0,s,b) += NUMSUBOPT(i,j-1,0,sLeft,bLeft);
            BPpartner(i,j,0,s,b)   = 0;  //no basepair partner of j in [i,j]
            LEVELleft(i,j,0,s,b)   = 0;
            LEVELright(i,j,0,s,b)  = 0;
            VISleft(i,j,0,s,b)     = sLeft;
            VISright(i,j,0,s,b)    = 0;
            BSleft(i,j,0,s,b)      = bLeft;
            BSright(i,j,0,s,b)     = 0;
            } // end if
          }// end for (bLeft=0;bLeft<BB;bLeft++)
    }// for (i=1;i<=n-THRESHOLD-1;i++)
#if 0
  numNativeStates = 0;
  for (s=0;s<SS;s++) 
    for (b=0;b<BB;b++) 
      numNativeStates += NUMSUBOPT(1,n,0,s,b); 
  printf("Number of Native States: %d\n\n",numNativeStates);
#endif
}


