//Header file for spectrum program
//P. Clote

//WARNING: YOU MUST have a directory DATA in the
//same directory as this program; otherwise there is
//a core dump

//WARNING -- GLOBAL variables are declared at the end of this header file

/*--------------------------------------------
  Headers 
--------------------------------------------*/
#define __SPECTRUM_H__
#include <stdio.h>
#include <string.h>
#include <ctype.h>      // character handling
#include <assert.h>
#include <limits.h>     //for INT_MAX
                                                                                
                                                                                


/*--------------------------------------------
  Constants
--------------------------------------------*/
#define DEBUG 0
#define PRINTENERGY 1
#define NN0 200
  //upper bound for N, where N is global variable = len(RNA)+3
#define NNover2 (NN/2)
#define KK0 50
#define KK ( MIN(KK0,(NN-THRESHOLD)/2) )
#define SS 16
#define SS0 65


#define A4 NN*KK*SS0*BB0
#define A3 KK*SS0*BB0
#define A2 SS0*BB0
#define A1 BB0
#define B4 NN*KK*SS*BB
#define B3 KK*SS*BB
#define B2 SS*BB
#define B1 BB


#define THRESHOLD 3
#define BB (THRESHOLD+2)
#define BB0 MAX(4,BB)


  //NN is upper bound for sequence length of RNA
  //Nucleotide sequence a1,...,an is stored by
  //array with values (n,a1,a2,...,an); ie length is a[0] in array
  //SS = 16 = 2**4 for encoding subsets of {A,C,G,U}
  //BB = threshold for encoding bitstring end visibility

#define INF INT_MAX/2
#define GUbond -1
#define AUbond -2
#define CGbond -3
        // for i and j to basepair, |j-i| > THRESHOLD
        // THRESHOLD must be >= 1 for code to work
#define MIN(x,y) ((x)<=(y) ? (x) : (y) )
#define MAX(x,y) ((x)>(y)  ? (x) : (y) )


/*--------------------------------------------
          Type definitions 
--------------------------------------------*/
//typedef short  MatrixType[NN][NN][KK][SS0][BB0];
//typedef float MatrixTypeDouble[NN][NN][KK][SS][BB];

typedef int INT;  //memory address
typedef float DOUBLE;  //memory address
typedef INT*  MatrixType;
typedef DOUBLE* MatrixTypeDouble;


/*--------------------------------------------
  Functional macros
--------------------------------------------*/
#define BIT(x,i)  ( ( (x)>>(i) ) & 1 )
#define in(x,s) (x=='A'?BIT(s,0):(x=='C'?BIT(s,1):(x=='G'?BIT(s,2):x=='U'&&BIT(s,3))))
        //determines whether nucleotide x (A,C,G,U) in set s


//#define CASE(i,j,k,s,b)        CASEfunction(M,i,j,k,s,b)
#define CASE(i,j,k,s,b)     CASEfunction(M,i,j,k,s,b)

//#define BP(i,j,k)              M[i][j][k][64][0]
#define BP(i,j,k)     M[A4*(i)+A3*(j)+A2*(k)+A1*64]

//#define BS(i,j,k)              M[i][j][k][64][1]
#define BS(i,j,k)     M[A4*(i)+A3*(j)+A2*(k)+A1*64+1]

//#define VIS(i,j,k)             M[i][j][k][64][2]
#define VIS(i,j,k)    M[A4*(i)+A3*(j)+A2*(k)+A1*64+2]

//#define MAXK(i,j)              M[i][j][0][64][3]
#define MAXK(i,j)     M[A4*(i)+A3*(j)+A2*0+A1*64+3]

//#define BPs(i,j,k,s,b)         M[i][j][k][s][b]
#define BPs(i,j,k,s,b)         M[A4*(i)+A3*(j)+A2*(k)+A1*(s)+b]

//#define LEVELleft(i,j,k,s,b)   M[j][i][k][s][b]
#define LEVELleft(i,j,k,s,b)   M[A4*(j)+A3*(i)+A2*(k)+A1*(s)+b]

//#define BPpartner(i,j,k,s,b)   M[i][j][k][16+(s)][b]
#define BPpartner(i,j,k,s,b)   M[A4*(i)+A3*(j)+A2*(k)+A1*(16+(s))+b]

//#define LEVELright(i,j,k,s,b)  M[j][i][k][16+(s)][b]
#define LEVELright(i,j,k,s,b)  M[A4*(j)+A3*(i)+A2*(k)+A1*(16+(s))+b]

//#define VISleft(i,j,k,s,b)     M[i][j][k][32+(s)][b] 
#define VISleft(i,j,k,s,b)     M[A4*(i)+A3*(j)+A2*(k)+A1*(32+(s))+b]

//#define VISright(i,j,k,s,b)  M[j][i][k][32+(s)][b]
#define VISright(i,j,k,s,b)    M[A4*(j)+A3*(i)+A2*(k)+A1*(32+(s))+b]

//#define BSleft(i,j,k,s,b)      M[i][j][k][48+(s)][b] 
#define BSleft(i,j,k,s,b)      M[A4*(i)+A3*(j)+A2*(k)+A1*(48+(s))+b]

//#define BSright(i,j,k,s,b)     M[j][i][k][48+(s)][b] 
#define BSright(i,j,k,s,b)     M[A4*(j)+A3*(i)+A2*(k)+A1*(48+(s))+b]

//#define NUMSUBOPT(i,j,k,s,b)   MM[i][j][k][s][b] 
#define NUMSUBOPT(i,j,k,s,b)   MM[B4*(i)+B3*(j)+B2*(k)+B1*(s)+b]


/*--------------------------------------------
NOTE:
  1) For 1<=i<=j<=n, and 0<=k<(j-i)/2 and 0 <= s < 16,
     M[i][j][k][s] is number of basepairs in s-visible, k-suboptimal
     sec str on [i,j], if such exist. If none exist and k=0, then
     initialize to 0; while if none exist and k>0, then initialize to -1. 
     Functional macro: BPs(i,j,k,s) 

  2) For 1<=i<=j<=n, and 0<=k<(j-i)/2 and s=48,
     M[i][j][k][48] is number of basepairs in k-suboptimal
     sec str on [i,j], if such exist. Same convention as previously
     if no basepairs exist.
     Functional macro: BP(i,j,k) 

  3) For 1<=i<=j<=n, and 0<=k<(j-i)/2 and 16 <= s < 31,
     M[i][j][k][s] is that value r, such that i<= r < j and
     r basepairs with j. If there is no such r, then initialized to 0.
     Functional macro: BPpartner(i,j,k,s) 

  4) For 1<=i<=j<=n, and 0<=k<(j-i)/2 and 0 <= s < 15,
     M[j][i][k][s] is kLeft, the suboptimal level of sec str 
     on [i,r-1], given a s-visible k-suboptimal sec str on [i,j].
     Functional macro: LEVELleft(i,j,k,s) 


  5) For 1<=i<=j<=n, and 0<=k<(j-i)/2 and 16 <= s < 31, 
     M[j][i][k][s] is kRight, the s-visible suboptimal level of sec str 
     on [r+1,j-1], given a s-visible k-suboptimal sec str on [i,j].
     Functional macro: LEVELright(i,j,k,s) 

  6) For 1<=i<=j<=n, and 0<=k<(j-i)/2 and s=48, 
     M[j][i][k][48] is s, the visibility of a (fixed, but arbitrary)
     k-suboptimal sec str on [i,j].
     Functional macro: VIS(i,j,k) 

NOTE: BS(i,j,k) = r means that none of the positions j-THRESHOLD,...,j-r
      is free, Thus r=0 means that nothing is visible while r=THRESHOLD+1
      means that all positions j-THRESHOLD,...,j are free.
--------------------------------------------*/





/*--------------------------------------------
          Prototypes
--------------------------------------------*/
//Functions from misc.c
void error(char *);
short *getRNA(char *flag, char *rna);
int guPair(char, char);
int auPair(char, char);
int cgPair(char, char);
int bondEnergy(int, int, char *);
int basePair(char, char);
int watsonCrick(char, char);
void printMatrix(MatrixType,int);
void printMatrixDouble(MatrixTypeDouble,int);
void printRNA(short *,short);            
int addElement(int x,int s);  //add x to set s  
int shiftAppendBitString(int b, int x);
int joinSandB(short *a, int s, int b, int i, int j);
int initializeBitString(int i,int j);
int initializeVis(short *a, int i,int j);
int canBasePairWithElementInSet(int x, int s);
char * constructString(short *a);
//int CASEfunction(MatrixType,int,int,int,int,int); 
//short CASEfunction(short *,short,short,short,short,short);

//Functions from energy.c
void initialize(MatrixType, MatrixTypeDouble, short*);  
void initializeBis(MatrixType, MatrixTypeDouble, short*);  
int computeEnergy(MatrixType, MatrixTypeDouble, short*);  
//void initializeParen(short [NN/2][NN],int);
//void backtrack(int,int,int,int,int,short [NN], short*,MatrixType);
void initializeParen(short**,int);
void backtrack(int,int,int,int,int,short*, short*,MatrixType);

//Functions from backtrack.c
//void createLevelSecStr(short [NN/2][NN],int ,short *,MatrixType);
void createLevelSecStr(short**,int ,short *,MatrixType);

//WARNING -- GLOBAL variables
/*--------------------------------------------
  Global variables
--------------------------------------------*/
MatrixType M;
MatrixTypeDouble MM;
int NN;
                                                                                
