
/*--------------------------------------------------------------
main.c
P. Clote, Aug 2003
--------------------------------------------------------------*/


#ifndef __SPECTRUM_H__
#include "spectrum.h"
#endif
                                                                                



main(int argc, char *argv[]) {
  // variable declarations
  FILE *file;
//  char  text[3*NN]; 
  char* text;
  char infileName[]  = "DATA/rnaSeqStrSubOptimum.input";
  char outfileName[] = "DATA/rnaSeqStrSubOptimum.output";

  float n0;
  int s,b;
  char *nuclSeq, *rnaSecStr;
  int i,n;   //i is loop control var, n is length of RNA sequence
  int k; //k takes values k=0,...,NN/2-1 for k-suboptimal
  int maxK;  //maximum k for interval 1...n
  int maxNumBP;
  char *flag, *rna;
  short *a; //a points to array a[0],...,a[n]
            //where a[0]=n, and a[1],...,a[n] are RNA nucleotides
            //and where n <= NN-2
//  short paren[NN/2][NN];  
  short **paren;
    //paren is parenthesis matrix for secondary structure
    //paren[k][1]...paren[k][n] is balanced parenthesis structure for
    //the k-suboptimal secondary structure S, ie number of basepairs in S
    //is k less than the number of basepairs in the optimal structure

  /**** Error handling for input file ****/  
  if (argc != 3) {
    fprintf(stderr,"%s -s/-f string/finame\n",argv[0]);
    fprintf(stderr,"\t-s: enter RNA string \n");
    fprintf(stderr,"\t-f: enter file of RNA in FASTA format\n");
    exit(1);
    }
  flag = argv[1];
  rna  = argv[2];
  a    = getRNA(flag,rna);
  n    = a[0];
  NN   = n+3;

  /**** Call calloc for arrays M,MM  ****/
  M  = (INT*) calloc(NN*NN*KK*SS0*BB0,sizeof(INT));
  if (M==NULL){
    printf("M is null\n");
    exit(1);
    }
  MM = (DOUBLE*) calloc(NN*NN*KK*SS*BB,sizeof(DOUBLE));
  if (MM==NULL){
    printf("MM is null\n");
    exit(1);
    }
  paren  = (short**) calloc(NNover2,sizeof(short*));
  for (i=0;i<NNover2;i++)
    paren[i]  = (short*) calloc(NN,sizeof(short));
  text   = (char*) calloc(3*NN,sizeof(char));
  
   
  initialize(M,MM,a);
  initializeBis(M,MM,a);
  if (DEBUG){
    printf("\nInitialization\n");
    printMatrix(M,n);
    }
  maxK = computeEnergy(M,MM,a);
  maxNumBP = BP(1,n,0); //max number of basepairs in opt structure
  if (!PRINTENERGY){
    printf("maxK: %d\n",maxK);
    printf("Max num basepairs: %d\n",BP(1,n,0));
    }
  if (DEBUG){
    printf("\nNum SubOptimal Computation\n");
    printMatrixDouble(MM,n);
    printf("\nEnergy Computation\n");
    printMatrix(M,n);
    }
  initializeParen(paren,n);
  createLevelSecStr(paren,maxK,a,M);
  for (k=0;k<=MIN(KK-1,maxK);k++){
    if (  BP(1,n,k) > 0 ){
      n0 = 0;
      for (s=0;s<SS;s++)
        for (b=0;b<BB;b++)
          if ( NUMSUBOPT(1,n,k,s,b)  > 0 ) { //Warning: some values are -1
            n0 += NUMSUBOPT(1,n,k,s,b);
            if (DEBUG)
              fprintf(stderr,"%d\t%d\t%d\t%f\t%d\n",k,s,b,NUMSUBOPT(1,n,k,s,b),BPs(1,n,k,s,b));
	  }
      if (PRINTENERGY){
        nuclSeq   = constructString(a);
        rnaSecStr = constructString(paren[k]);
        file = fopen(infileName,"w");
        sprintf(text,">k: %d has %d basepairs\n",k,BP(1,n,k));
        fprintf(file,text);
        sprintf(text,">Num %d-suboptimal sec str: %.0lf\n",k,n0);
        fprintf(file,text);
        fprintf(file,"%s\n%s\n@",nuclSeq,rnaSecStr);
        fclose(file);
        sprintf(text,"RNAeval < %s\n",infileName);
        system(text);
        }
      else{
        printf("k: %d has %d basepairs\n",k,BP(1,n,k));
        printf("Num %d-suboptimal sec str: %.0lf\n",k,n0);
        printRNA(a,0);
        printRNA(paren[k],1);
        }
      }
    }
#if 0
#endif
} 


